__version__ = "0.0.4"

META_DATA = {
    "author": "Omar Shouman",
    "author_email": "o.shouman@tum.de",
    "description": "Deep Learning for Proteomics",
    "package_name": "DLOmix",
    "copyright_text": "2022, Wilhelm Lab, TU Munich.",
    "github_url": "https://github.com/wilhelm-lab/dlomix",
}
